#include <GnugoOpponent.h>

GnugoOpponent::GnugoOpponent(SgfNode* root, bool black, int level, double komi)
	:Opponent(), m_gnugo_pid(-1), m_count(0), m_black(black), m_last_move(0)
{
	std::vector<std::string> argv;
	argv.push_back("gnugo");
	argv.push_back("--mode");
	argv.push_back("gtp");
	Glib::spawn_async_with_pipes(Glib::get_current_dir(),argv,
	                             Glib::SPAWN_SEARCH_PATH,sigc::slot<void>(),
	                             &m_gnugo_pid,&m_gnugo_fd_stdin,
	                             &m_gnugo_fd_stdout);
	m_gnugo_stdin=Glib::IOChannel::create_from_fd(m_gnugo_fd_stdin);
	m_gnugo_stdout=Glib::IOChannel::create_from_fd(m_gnugo_fd_stdout);
	m_conn_handle=Glib::signal_io().connect(sigc::mem_fun(this,&GnugoOpponent::process_input),m_gnugo_stdout,Glib::IO_IN);
	list<SgfProperty*>::const_iterator iter=root->Begin();
	m_last_move=root;
	while(iter!=root->End()){
		send_instruction(*iter);
		iter++;
	}
        m_gnugo_stdin->write(ustring::compose("%1 level %2\n",++m_count,level));
        Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 level %2\n",m_count,level)<<"]"<<endl;
        m_gnugo_queue.push(t_message(m_count,LEVEL));
        m_gnugo_stdin->flush();
        m_gnugo_stdin->write(ustring::compose("%1 komi %2\n",++m_count,komi));
        Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 komi %2\n",m_count,komi)<<"]"<<endl;
        m_gnugo_queue.push(t_message(m_count,KOMI));
        m_gnugo_stdin->flush();

}

GnugoOpponent::~GnugoOpponent()
{
    m_conn_handle.disconnect();
    m_gnugo_stdin->write("quit\n");
    m_gnugo_stdin->flush();
    m_gnugo_stdin->close();
    m_gnugo_stdout->close();
    Glib::spawn_close_pid(m_gnugo_pid);
}

void GnugoOpponent::ToPlay(SgfNode* lastmove)
{
    if(!m_is_playing){
        if(!lastmove) send_instruction(PASS);
        else{
            list<SgfProperty*>::const_iterator iter=lastmove->Begin();
            while(iter!=lastmove->End()){
                    send_instruction(*iter);
                    iter++;
            }
            send_instruction(GENMOVE);
            m_is_playing=true;
        }
    }
    m_last_move=lastmove;
}

void GnugoOpponent::GetScore()
{
    m_is_playing=true;
    send_instruction(SCORE);
    send_instruction(STATUS_LIST_ALIVE);
    send_instruction(STATUS_LIST_DEAD);
    send_instruction(STATUS_LIST_SEKI);
    send_instruction(STATUS_LIST_BLACK_TERRITORY);
    send_instruction(STATUS_LIST_WHITE_TERRITORY);
    send_instruction(STATUS_LIST_DAME);
}

void GnugoOpponent::Resignation()
{
}

Coord GnugoOpponent::gnugo_to_coord(string coords)
{
    char acoord(0);
    int xcoord(-1),ycoord(-1);
    acoord=coords[0];
    istringstream strin2(coords.substr(1));
    strin2>>ycoord;
    if(strin2.fail()) throw GnugoError(ustring::compose("Unexpected return. string was \"%1\"",coords));
    if(acoord > 'H') --acoord; //'I' is skipped in output.
    xcoord=int(acoord)-int('A');
    return Coord(xcoord,ycoord-1);
}

void GnugoOpponent::check_status_emit()
{
    if(m_final_dead.size()>=1 && m_final_black_territory.size()>=1 && m_final_white_territory.size()>=1){
        m_final_dead.pop_back();
        m_final_black_territory.pop_back();
        m_final_white_territory.pop_back();
        m_signal_final_state.emit(m_final_dead,m_final_black_territory,m_final_white_territory);
    }
}

bool GnugoOpponent::process_input(Glib::IOCondition io_cond)
{
    if((io_cond & Glib::IO_IN) == 0){
        Logger::log()<<"GnugoOpponent::process_input: bad gnugo response."<<endl;
        return true;
    }
    if(!m_conn_handle.connected())
    {
        Logger::log()<<"GnugoOpponent::process_input: disconnected"<<endl;
        return true;
    }
    ustring buf;
    m_gnugo_stdout->read_line(buf);
    if(buf=="\n"){
        Logger::log()<<"[gnugo->client"<<endl<<m_gnugo_stdout_buffer<<"]"<<endl;
        if(m_gnugo_stdout_buffer.size()==0){
            Logger::log()<<"empty gnugo response, weird"<<endl;
        } else if(m_gnugo_stdout_buffer[0]=='?' || m_gnugo_stdout_buffer[0]=='='){
            istringstream strin(m_gnugo_stdout_buffer.substr(1));
            int num;
            strin>>num;
            if(strin.fail())
                throw GnugoError(ustring::compose("Unexpected return. string was \"%1\"",m_gnugo_stdout_buffer));
            if(m_gnugo_queue.front().first != num)
                throw GnugoError(ustring::compose("Expected return for instruction %1, recieved for %2",m_gnugo_queue.front().first,num));
            if(m_gnugo_stdout_buffer[0]=='?'){
                Logger::log()<<"Warning, gnugo returned an error: \""<<m_gnugo_stdout_buffer<<"\""<<endl;
                Logger::log()<<"Ask for gnugo board state."<<endl;
                send_instruction(SHOWBOARD);
            }else{
                switch(m_gnugo_queue.front().second){
                    case BOARDSIZE:
                        break;
                    case PLAYBLACK:
                        break;
                    case PLAYWHITE:
                        break;
                    case GENMOVE:{
                        string strcoords;
                        strin>>strcoords;
                        if(strcoords.size()<2) throw GnugoError(ustring::compose("Unexpected return. string was \"%1\"",strcoords));
                        else if(strcoords=="PASS"){
                            SgfNode* node=new SgfNode(m_last_move);
                            int size=m_last_move->GetBoard().Size();
                            if(m_black) node->AddProperty(SgfProperty::B,ustring(Coord(size,size)));
                            else node->AddProperty(SgfProperty::W,ustring(Coord(size,size)));
                            m_is_playing=false;
                            m_signal_played.emit(node);
                        } else if(strcoords=="resign"){
                            m_is_playing=false;
                            m_signal_resign.emit();
                        } else if(Glib::Ascii::isalpha(strcoords[0]) && Glib::Ascii::isupper(strcoords[0])){
                            Coord coords=gnugo_to_coord(strcoords);
                            SgfNode* node=new SgfNode(m_last_move);
                            if(m_black) node->AddProperty(SgfProperty::B,ustring(coords));
                            else node->AddProperty(SgfProperty::W,ustring(coords));
                            m_is_playing=false;
                            //debug
                            send_instruction(SHOWBOARD);
                            m_signal_played.emit(node);
                        } else throw GnugoError(ustring::compose("Unexpected return. string was \"%1\"",strcoords));
                    }
                        break;
                    case STATUS_LIST_BLACK_TERRITORY:
                        while(!strin.eof()){
                            string strcoords;
                            strin>>strcoords;
                            if(strcoords.size()>=2) m_final_black_territory.push_back(gnugo_to_coord(strcoords));
                        }
                        m_final_black_territory.push_back(Coord(-1,-1));
                        check_status_emit();
                        break;
                    case STATUS_LIST_WHITE_TERRITORY:
                        while(!strin.eof()){
                            string strcoords;
                            strin>>strcoords;
                            if(strcoords.size()>=2) m_final_white_territory.push_back(gnugo_to_coord(strcoords));
                        }
                        m_final_white_territory.push_back(Coord(-1,-1));
                        check_status_emit();
                        break;
                    case STATUS_LIST_DEAD:
                        while(!strin.eof()){
                            string strcoords;
                            strin>>strcoords;
                            if(strcoords.size()>=2) m_final_dead.push_back(gnugo_to_coord(strcoords));
                        }
                        m_final_dead.push_back(Coord(-1,-1));
                        check_status_emit();
                        break;
                    default:
                        break;
                }
            }
            m_gnugo_queue.pop();
        } else throw GnugoError(ustring::compose("Unexpected return. string was \"%1\"",m_gnugo_stdout_buffer));
        m_gnugo_stdout_buffer.clear();
    } else {
        m_gnugo_stdout_buffer.append(buf);
    }
    return true;
}

void GnugoOpponent::send_instruction(const SgfProperty* prop)
{
    if(prop){
        switch(prop->PropIdent()){
            case SgfProperty::SZ:{
                m_gnugo_stdin->write(ustring::compose("%1 boardsize %2\n",++m_count,prop->Value2I()));
                Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 boardsize %2\n",m_count,prop->Value2I())<<"]"<<endl;;
                m_gnugo_queue.push(t_message(m_count,BOARDSIZE));
            } break;
            case SgfProperty::B:{
                Coord c=prop->Value2C();
                int size=m_last_move->GetBoard().Size();
                if(c.i()==size){
                    m_gnugo_stdin->write(ustring::compose("%1 play black PASS\n",++m_count));
                    Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 play black PASS\n",m_count)<<"]"<<endl;
                } else {
                    char xcoord=char(int('A')+c.i());
                    if(xcoord>='I') ++xcoord;
                    m_gnugo_stdin->write(ustring::compose("%1 play black %2%3\n",++m_count,xcoord,c.j()+1));
                    Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 play black %2%3\n",m_count,xcoord,c.j()+1)<<"]"<<endl;
                }
                m_gnugo_queue.push(t_message(m_count,PLAYBLACK));
            } break;
            case SgfProperty::W:{
                Coord c=prop->Value2C();
                int size=m_last_move->GetBoard().Size();
                if(c.i()==size){
                    m_gnugo_stdin->write(ustring::compose("%1 play white PASS\n",++m_count));
                    Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 play white PASS\n",m_count)<<"]"<<endl;
                } else {
                    char xcoord=char(int('A')+c.i());
                    if(xcoord>='I') ++xcoord;
                    m_gnugo_stdin->write(ustring::compose("%1 play white %2%3\n",++m_count,xcoord,c.j()+1));
                    Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 play white %2%3\n",m_count,xcoord,c.j()+1)<<"]"<<endl;
                }
                m_gnugo_queue.push(t_message(m_count,PLAYWHITE));
            } break;
        }
    }
    m_gnugo_stdin->flush();
}

void GnugoOpponent::send_instruction(GnugoMessage message)
{
    switch(message){
        case SHOWBOARD:
            m_gnugo_stdin->write(ustring::compose("%1 showboard\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 showboard\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,SHOWBOARD));
            break;
        case GENMOVE:
            if(m_black){
                    m_gnugo_stdin->write(ustring::compose("%1 genmove black\n",++m_count));
                    Logger::log()<<"[client->gnugo"<<endl<<ustring::compose("%1 genmove black\n",m_count)<<"]"<<endl;
            }else{
                    m_gnugo_stdin->write(ustring::compose("%1 genmove white\n",++m_count));
                    Logger::log()<<"[client->gnugo"<<endl<<ustring::compose("%1 genmove white\n",m_count)<<"]"<<endl;
            }
            m_gnugo_queue.push(t_message(m_count,GENMOVE));
            break;
        case PASS:
            m_gnugo_stdin->write(ustring::compose("%1 pass\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 pass\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,SHOWBOARD));
            break;
        case SCORE:
            m_gnugo_stdin->write(ustring::compose("%1 final_score\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_score\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,SCORE));
            break;
        case STATUS_LIST_ALIVE:
            m_gnugo_stdin->write(ustring::compose("%1 final_status_list alive\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_status_list alive\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,STATUS_LIST_ALIVE));
            break;
        case STATUS_LIST_DEAD:
            m_gnugo_stdin->write(ustring::compose("%1 final_status_list dead\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_status_list dead\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,STATUS_LIST_DEAD));
            break;
        case STATUS_LIST_SEKI:
            m_gnugo_stdin->write(ustring::compose("%1 final_status_list seki\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_status_list seki\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,STATUS_LIST_SEKI));
            break;
        case STATUS_LIST_BLACK_TERRITORY:
            m_gnugo_stdin->write(ustring::compose("%1 final_status_list black_territory\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_status_list black_territory\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,STATUS_LIST_BLACK_TERRITORY));
            break;
        case STATUS_LIST_WHITE_TERRITORY:
            m_gnugo_stdin->write(ustring::compose("%1 final_status_list white_territory\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_status_list white_territory\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,STATUS_LIST_WHITE_TERRITORY));
            break;
        case STATUS_LIST_DAME:
            m_gnugo_stdin->write(ustring::compose("%1 final_status_list dame\n",++m_count));
            Logger::log()<<"[client->gnugo:"<<endl<<ustring::compose("%1 final_status_list dame\n",m_count)<<"]"<<endl;
            m_gnugo_queue.push(t_message(m_count,STATUS_LIST_DAME));
            break;
    }
    m_gnugo_stdin->flush();
}

ostream& operator<<(ostream& os,const GnugoError& err)
{
    if(err.Fatal())
        os<<"Fatal: ";
    os<<"GnugoError: "<<err.what();
    return os;
}
