#ifndef _GNUGOOPPONENT_H
#define _GNUGOOPPONENT_H

#include "Opponent.h"
#include <queue>
#include "Logger.h"

class GnugoOpponent: public Opponent {
    public:
        GnugoOpponent(SgfNode* root, bool black, int level=10, double komi=5.5);
        virtual ~GnugoOpponent();
        virtual void ToPlay(SgfNode* lastmove);
        virtual void GetScore();
        virtual void Resignation();
    private:
        enum GnugoMessage{
            GENMOVE,
            PLAYBLACK,
            PLAYWHITE,
            BOARDSIZE,
            SHOWBOARD,
            PASS,
            LEVEL,
            SCORE,
            KOMI,
            STATUS_LIST_ALIVE,
            STATUS_LIST_DEAD,
            STATUS_LIST_SEKI,
            STATUS_LIST_WHITE_TERRITORY,
            STATUS_LIST_BLACK_TERRITORY,
            STATUS_LIST_DAME
        };
        vector<Coord> m_final_black_territory;
        vector<Coord> m_final_white_territory;
        vector<Coord> m_final_dead;
        typedef std::pair<int,GnugoMessage> t_message;
        bool process_input(Glib::IOCondition io_cond);
        void send_instruction(const SgfProperty* prop);
        void send_instruction(GnugoMessage message);
        void check_status_emit();
        void ask_for_move();
        Coord gnugo_to_coord(string coords);
        Glib::Pid m_gnugo_pid;
        ustring m_gnugo_stdout_buffer;
        Glib::RefPtr<Glib::IOChannel> m_gnugo_stdin;
        int							 m_gnugo_fd_stdin;
        Glib::RefPtr<Glib::IOChannel> m_gnugo_stdout;
        int							 m_gnugo_fd_stdout;
        sigc::connection m_conn_handle;
        int	m_count;
        bool m_black;
        SgfNode* m_last_move;
        std::queue<std::pair<int,GnugoMessage> > m_gnugo_queue;
};

class GnugoError: public Error{
    public:
        GnugoError(const ustring& err, bool fatal=false, int code=0):Error(err,fatal,code) {}
        friend ostream& operator<<(ostream& os, const GnugoError& err);
};

#endif /*_GNUGOOPPONENT_H*/
