#include "GobanWidget.h"

GobanWidget::GobanWidget(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& refBuilder)
    :m_board(0), m_cur_x(-1), m_cur_y(-1), m_cur_col(BLACK), m_line_space(0), m_draw_ticks(false),
    m_refGdkWindow(0), Glib::ObjectBase("GobanWidget"), Gtk::Widget(cobject)
{
    set_flags(Gtk::NO_WINDOW | Gtk::CAN_FOCUS);
    signal_button_press_event().connect
        (
         sigc::mem_fun(*this,&GobanWidget::on_button_press_event)
        );
    signal_motion_notify_event().connect
        (
         sigc::mem_fun(*this,&GobanWidget::on_motion_notify_event)
        );
}

GobanWidget::~GobanWidget()
{
    m_refGdkWindow->reference();
    m_refGdkWindow->clear();
}

void GobanWidget::SetBoard(const Board* board)
{
    Gtk::Allocation allocation=get_allocation();
    if(board){
        if(!m_board && is_realized()){
            m_board=board;
            if (allocation.get_width()<allocation.get_height())
                m_line_space=(allocation.get_width())/(m_board->Size()+1);
            else
                m_line_space=(allocation.get_height())/(m_board->Size()+1);
            m_goban_map = Gdk::Pixmap::create(m_refGdkWindow,allocation.get_width(),allocation.get_height());
            load_pixbuffers();
            draw_goban_pixmap();
        } else {
            m_board=board;
            draw_goban_pixmap();
        }
    } else {
        m_board=0;
        m_cur_x=-1;
        m_cur_y=-1;
        m_cur_col=BLACK;
        m_line_space=0;
    }
    if(is_realized())
        queue_draw();
}

void GobanWidget::on_realize()
{
    //cout<<"start realize"<<endl;
    Gtk::Widget::on_realize();
    if(!m_refGdkWindow){
        GdkWindowAttr attributes;
        memset(&attributes, 0, sizeof(attributes));

        Gtk::Allocation allocation = get_allocation();

        //Set initial position and size of the Gdk::Window:
        attributes.x = allocation.get_x();
        attributes.y = allocation.get_y();
        attributes.width = allocation.get_width();
        attributes.height = allocation.get_height();

        attributes.event_mask = get_events () | 
            GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
            GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
            GDK_POINTER_MOTION_HINT_MASK | GDK_LEAVE_NOTIFY_MASK;
        attributes.window_type = GDK_WINDOW_CHILD;
        attributes.wclass = GDK_INPUT_OUTPUT;

        m_refGdkWindow = Gdk::Window::create(get_window() /*parent*/ , &attributes,
                GDK_WA_X | GDK_WA_Y);
        m_refGdkWindow->reference();
        if(allocation.get_width()!=0 && allocation.get_height()!=0)
            m_goban_map = Gdk::Pixmap::create(m_refGdkWindow,allocation.get_width(),allocation.get_height());
        unset_flags(Gtk::NO_WINDOW);
        set_window(m_refGdkWindow);
        m_refGdkWindow->set_user_data(gobj());
        if(m_board){
            if (allocation.get_width()<allocation.get_height()){
                m_line_space=(allocation.get_width())/(m_board->Size()+1);
            } else {
                m_line_space=(allocation.get_height())/(m_board->Size()+1);
            }
            if(allocation.get_width()!=0 && allocation.get_height()!=0){
                load_pixbuffers();
                draw_goban_pixmap();
            }
        }
        m_logo=Gdk::Pixbuf::create_from_file(DATADIR"logo.svg",300,-1);
    }
}

void GobanWidget::load_pixbuffers()
{
    m_board_pix=Gdk::Pixbuf::create_from_file(DATADIR"goban.svg",
                                m_line_space*(m_board->Size()+1),
                                m_line_space*(m_board->Size()+1));
    m_stone_black=Gdk::Pixbuf::create_from_file(DATADIR"black.svg",
                                (int)(0.98*m_line_space),
                                (int)(0.98*m_line_space));
    m_stone_white=Gdk::Pixbuf::create_from_file(DATADIR"white.svg",
                                (int)(0.98*m_line_space),
                                (int)(0.98*m_line_space));
    m_stone_black_cursor=Gdk::Pixbuf::create_from_file(DATADIR"black_moving.svg",
                                (int)(0.98*m_line_space),
                                (int)(0.98*m_line_space));
    m_stone_white_cursor=Gdk::Pixbuf::create_from_file(DATADIR"white_moving.svg",
                                (int)(0.98*m_line_space),
                                (int)(0.98*m_line_space));
    m_void_cursor=Gdk::Pixbuf::create_from_file(DATADIR"void.svg",
                                (int)(0.98*m_line_space),
                                (int)(0.98*m_line_space));
    m_shadow=Gdk::Pixbuf::create_from_file(DATADIR"shadow.svg",
                                (int)(1.1*m_line_space),
                                (int)(1.1*m_line_space));
}

void GobanWidget::on_unrealize()
{
    m_refGdkWindow.clear();
    m_goban_map.clear();
    Gtk::Widget::on_unrealize();
}

void GobanWidget::on_map()
{
  Gtk::Widget::on_map();
}

void GobanWidget::on_unmap()
{
  Gtk::Widget::on_unmap();
}

void GobanWidget::on_size_request(Gtk::Requisition* requisition)
{
    //cout<<"on size requisition"<<endl;
    *requisition = Gtk::Requisition();
    requisition->height = 200;
    requisition->width = 200;
    //cout<<"leave on_size_request"<<endl;
}

void GobanWidget::on_size_allocate(Gtk::Allocation& allocation)
{
    //cout<<"on_size_allocate"<<endl;
    set_allocation(allocation);
    if(m_refGdkWindow) {
        m_refGdkWindow->move_resize(allocation.get_x(), allocation.get_y(),
                                    allocation.get_width(), allocation.get_height());
        if(allocation.get_width()!=0 && allocation.get_height()!=0) 
            m_goban_map = Gdk::Pixmap::create(m_refGdkWindow,allocation.get_width(),allocation.get_height());
    }
    if(m_board){
        if (allocation.get_width()<allocation.get_height()){
            m_line_space=(allocation.get_width())/(m_board->Size()+1);
        } else {
            m_line_space=(allocation.get_height())/(m_board->Size()+1);
        }
        if(allocation.get_width()!=0 && allocation.get_height()!=0) load_pixbuffers();
        draw_goban_pixmap();
    }
    //cout<<"leave on_size_allocate"<<endl;
}

void GobanWidget::top_left_corner(int& x, int& y)
{
    int xc,yc;

    Gtk::Allocation allocation = get_allocation();

    xc=allocation.get_width()/2;
    yc=allocation.get_height()/2;
    if (m_board->Size()%2==0) {
            x=xc-((m_board->Size())/2-0.5)*m_line_space;
            y=yc-((m_board->Size())/2-0.5)*m_line_space;
    } else{
            x=xc-((m_board->Size()-1)/2)*m_line_space;
            y=yc-((m_board->Size()-1)/2)*m_line_space;
    }
}

void GobanWidget::draw_goban_pixmap()
{
    Cairo::RefPtr<Cairo::Context> cr = m_goban_map->create_cairo_context();
    Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(m_goban_map);
    int w,h;
    m_goban_map->get_size(w,h);
    cr->rectangle(0, 0, w, h);
    Gdk::Cairo::set_source_color(cr, get_style()->get_bg(Gtk::STATE_NORMAL));
    cr->clip_preserve();
    cr->fill();
    gc->set_clip_rectangle(Gdk::Rectangle(0,0,w,h));
    draw_goban_wood(gc,cr);
    draw_labels(gc,cr,false);
    draw_stones(gc,cr);
    draw_labels(gc,cr,true);
    draw_ko(gc,cr);
    draw_territory(gc,cr);
}

bool GobanWidget::on_expose_event(GdkEventExpose* event)
{
    if(!is_realized()) return false;
    Cairo::RefPtr<Cairo::Context> cr = m_refGdkWindow->create_cairo_context();
    Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(get_window());
    if(!m_board) {
        cr->set_source_rgb(1,1,1);
        if(event){
            Gdk::Cairo::set_source_color(cr, get_style()->get_bg(Gtk::STATE_NORMAL));
            cr->paint();
            cr->rectangle(event->area.x, event->area.y, event->area.width, event->area.height);
            gc->set_clip_rectangle(Gdk::Rectangle(event->area.x, event->area.y, event->area.width, event->area.height));
            cr->clip_preserve();
            cr->fill();
            get_window()->draw_pixbuf(gc,m_logo,0,0,
                    (get_allocation().get_width()-m_logo->get_width())/2,
                     (get_allocation().get_height()-m_logo->get_height())/2,
                     -1,-1,Gdk::RGB_DITHER_NORMAL,0,0);
        }
    } else {
        if(event){
            gc->set_clip_rectangle(Gdk::Rectangle(event->area.x,event->area.y,event->area.width,event->area.height));
        }
        get_window()->draw_drawable(gc,m_goban_map,0,0,0,0);
    } 
    return false;
}

void GobanWidget::draw_goban_wood(Glib::RefPtr<Gdk::GC> gc, Cairo::RefPtr<Cairo::Context> cr, int x, int y, int w, int h)
{
    cr->save();
    if(x==-1){
        x=y=0;
        w=h=m_board->Size();
    }
    int x0,y0,i;
    top_left_corner(x0,y0);

    Glib::RefPtr<Pango::Layout> lt=Pango::Layout::create(cr);
    ustring fontstring=ustring::compose("sans %1", (int)(0.25*m_line_space));
    Pango::FontDescription font(fontstring);
    lt->set_font_description(font);

    //goban wood:
    m_goban_map->draw_pixbuf(gc,m_board_pix,0,0,x0-m_line_space,y0-m_line_space,-1,-1,Gdk::RGB_DITHER_NORMAL,0,0);

    //lines
    int line_width=int(0.04*m_line_space);
    if(line_width%2) line_width-=1;
    if(!line_width) line_width=2;
    cr->set_line_width(line_width);
    for (i=x;i<x+w;++i){
        cr->move_to(x0+i*m_line_space,y0);
        cr->line_to(x0+i*m_line_space,y0+(m_board->Size()-1)*m_line_space);
    }
    for (i=y;i<y+h;++i){
        cr->move_to(x0,y0+i*m_line_space);
        cr->line_to(x0+(m_board->Size()-1)*m_line_space,y0+i*m_line_space);
    }
    cr->rectangle(x0,y0,m_line_space*(m_board->Size()-1),m_line_space*(m_board->Size()-1));
    cr->set_source_rgb(0,0,0);
    cr->stroke();

    if(m_draw_ticks){
        for (i=x;i<x+w;++i){
            char letter=(char)((int)'A'+i);
            if(letter>='I') ++letter;
            cr->move_to(x0+(i-0.10)*m_line_space,y0-0.85*m_line_space);
            lt->update_from_cairo_context(cr);
            lt->set_text(ustring::compose("%1",letter));
            lt->add_to_cairo_context(cr);
            cr->move_to(x0+(i-0.10)*m_line_space,y0+(m_board->Size()-0.55)*m_line_space);
            lt->update_from_cairo_context(cr);
            lt->set_text(ustring::compose("%1",letter));
            lt->add_to_cairo_context(cr);
        }
        for (i=y;i<y+h;++i){
            if(m_board->Size()-i>9) cr->move_to(x0-0.88*m_line_space,y0+(i-0.2)*m_line_space);
            else cr->move_to(x0-0.88*m_line_space,y0+(i-0.2)*m_line_space);
            lt->update_from_cairo_context(cr);
            lt->set_text(ustring::compose("%1",m_board->Size()-i));
            lt->add_to_cairo_context(cr);
            if(m_board->Size()-i>9) cr->move_to(x0+(m_board->Size()-0.58)*m_line_space,y0+(i-0.2)*m_line_space);
            else cr->move_to(x0+(m_board->Size()-0.43)*m_line_space,y0+(i-0.2)*m_line_space);
            lt->update_from_cairo_context(cr);
            lt->set_text(ustring::compose("%1",m_board->Size()-i));
            lt->add_to_cairo_context(cr);
        }
    }
    cr->fill();
    /*hochis*/
    if(m_board->Size()==9){
        cr->set_source_rgb(0,0,0);
        cr->arc(x0+4*m_line_space, y0+4*m_line_space, 0.1*m_line_space,0.,2*M_PI);
    }
    else if(m_board->Size()==13){
        cr->set_source_rgb(0,0,0);
        cr->arc(x0+3*m_line_space, y0+3*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+3*m_line_space, y0+9*m_line_space);
        cr->arc(x0+3*m_line_space, y0+9*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+9*m_line_space, y0+3*m_line_space);
        cr->arc(x0+9*m_line_space, y0+3*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+9*m_line_space, y0+9*m_line_space);
        cr->arc(x0+9*m_line_space, y0+9*m_line_space, 0.1*m_line_space,0.,2*M_PI);
    } else if(m_board->Size()==19){
        cr->set_source_rgb(0,0,0);
        cr->arc(x0+3*m_line_space, y0+3*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+3*m_line_space, y0+9*m_line_space);
        cr->arc(x0+3*m_line_space, y0+9*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+3*m_line_space, y0+15*m_line_space);
        cr->arc(x0+3*m_line_space, y0+15*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+9*m_line_space, y0+3*m_line_space);
        cr->arc(x0+9*m_line_space, y0+3*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+9*m_line_space, y0+9*m_line_space);
        cr->arc(x0+9*m_line_space, y0+9*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+9*m_line_space, y0+15*m_line_space);
        cr->arc(x0+9*m_line_space, y0+15*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+15*m_line_space, y0+3*m_line_space);
        cr->arc(x0+15*m_line_space, y0+3*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+15*m_line_space, y0+9*m_line_space);
        cr->arc(x0+15*m_line_space, y0+9*m_line_space, 0.1*m_line_space,0.,2*M_PI);
        cr->move_to(x0+15*m_line_space, y0+15*m_line_space);
        cr->arc(x0+15*m_line_space, y0+15*m_line_space, 0.1*m_line_space,0.,2*M_PI);
    }
    cr->fill();
    cr->restore();
}

void GobanWidget::draw_stones(Glib::RefPtr<Gdk::GC> gc, Cairo::RefPtr<Cairo::Context> cr, int x, int y, int w, int h)
{
        /*stones*/
    if(x==-1){
        x=y=0;
        w=h=m_board->Size();
    }
    int x0,y0,i,j;
    top_left_corner(x0,y0);

    for (j=y; j < y+h; ++j){
        for (i=x; i<x+w; ++i){
            if((*m_board)(i,j)!=EMPTY)
                m_goban_map->draw_pixbuf(gc,m_shadow,
                        0,0,
                        x0+(i-0.6)*m_line_space,y0+(j-0.4)*m_line_space,
                        -1,-1,
                        Gdk::RGB_DITHER_NORMAL,
                        0,0);
        }
    }

    for (j=y; j < y+h; ++j){
        for (i=x; i<x+w; ++i){
            //try{
            if ((*m_board)(i,j)==EMPTY) continue;
            if ((*m_board)(i,j)==BLACK){
                m_goban_map->draw_pixbuf(gc,m_stone_black,
                        0,0,
                        x0+(i-0.49)*m_line_space,y0+(j-0.49)*m_line_space,
                        -1,-1,
                        Gdk::RGB_DITHER_NORMAL,
                        0,0);
            }else{
                m_goban_map->draw_pixbuf(gc,m_stone_white,
                        0,0,
                        x0+(i-0.49)*m_line_space,y0+(j-0.49)*m_line_space,
                        -1,-1,
                        Gdk::RGB_DITHER_NORMAL,
                        0,0);
            }
            //} catch(BoardError& err){
            //    cout<<err<<endl;
            //}
        }
    }
}

void GobanWidget::draw_labels(Glib::RefPtr<Gdk::GC> gc,Cairo::RefPtr<Cairo::Context> cr, bool stone_labels)
{
    int x0,y0;
    top_left_corner(x0,y0);
    Glib::RefPtr<Pango::Layout> lt=Pango::Layout::create(cr);
    ustring fontstring=ustring::compose("sans %1", (int)(0.65*m_line_space));
    Pango::FontDescription font(fontstring);
    lt->set_font_description(font);
    list<Label>::const_iterator iter=m_board->LabelBegin();
    while(iter!=m_board->LabelEnd()){
        cr->save();
        if(stone_labels){
            if((*m_board)(iter->Coordinates())==EMPTY){
                iter++;
                continue;
            }
        } else {
            if((*m_board)(iter->Coordinates())!=EMPTY){
                iter++;
                continue;
            }
        }
        if((*m_board)(iter->Coordinates())==EMPTY || (*m_board)(iter->Coordinates())==WHITE)
            cr->set_source_rgb(0,0,0);
        else
            cr->set_source_rgb(1,1,1);
        if((*m_board)(iter->Coordinates())==EMPTY){
            m_goban_map->draw_pixbuf(gc,m_board_pix,
                    (iter->Coordinates().i()+0.5)*m_line_space,
                    (iter->Coordinates().j()+0.5)*m_line_space,
                    x0+(iter->Coordinates().i()-0.5)*m_line_space,
                    y0+(iter->Coordinates().j()-0.5)*m_line_space,
                    m_line_space,
                    m_line_space,
                    Gdk::RGB_DITHER_NORMAL,0,0);
        }
        if(iter->GetType()==Label::LETTER){
            cr->move_to(x0+(iter->Coordinates().i()-0.3)*m_line_space,y0+(iter->Coordinates().j()-0.55)*m_line_space);
            lt->update_from_cairo_context(cr);
            lt->set_text(ustring::compose("%1",iter->Letter()));
            lt->add_to_cairo_context(cr);
            cr->fill();
        } else if(iter->GetType()==Label::CIRCLE){
            cr->move_to(x0+(iter->Coordinates().i()+0.3)*m_line_space,y0+iter->Coordinates().j()*m_line_space);
            cr->arc(x0+iter->Coordinates().i()*m_line_space,y0+iter->Coordinates().j()*m_line_space,0.3*m_line_space,0, 2*M_PI);
            cr->stroke();
        } else if(iter->GetType()==Label::SQUARE){
            cr->move_to(x0+(iter->Coordinates().i()-0.3)*m_line_space,y0+(iter->Coordinates().j()-0.3)*m_line_space);
            cr->rel_line_to(0.6*m_line_space,0);
            cr->rel_line_to(0,0.6*m_line_space);
            cr->rel_line_to(-0.6*m_line_space,0);
            cr->rel_line_to(0,-0.6*m_line_space);
            cr->stroke();
        } else if(iter->GetType()==Label::TRIANGLE){
            cr->move_to(x0+(iter->Coordinates().i())*m_line_space,y0+(iter->Coordinates().j()-0.3)*m_line_space);
            cr->rel_line_to(sqrt(3.0)*0.3*m_line_space*0.5,1.75*0.3*m_line_space);
            cr->rel_line_to(-2*sqrt(3.0)*0.3*m_line_space*0.5,0);
            cr->rel_line_to(sqrt(3.0)*0.3*m_line_space*0.5,-1.75*0.3*m_line_space);
            cr->stroke();
        }
        cr->restore();
        ++iter;
    }
}

void GobanWidget::draw_territory(Glib::RefPtr<Gdk::GC> gc,Cairo::RefPtr<Cairo::Context> cr)
{
    int x0,y0;
    top_left_corner(x0,y0);
    list<Coord>::const_iterator iter=m_board->BlackTerritoryBegin();
    while(iter!=m_board->BlackTerritoryEnd()){
        cr->save();
        cr->set_source_rgb(1,1,1);
        cr->move_to(x0+(iter->i()+0.15)*m_line_space,y0+iter->j()*m_line_space);
        cr->arc(x0+iter->i()*m_line_space,y0+iter->j()*m_line_space,0.15*m_line_space,0, 2*M_PI);
        cr->stroke_preserve();
        cr->set_source_rgb(0,0,0);
        cr->fill();
        cr->restore();
        ++iter;
    }
    iter=m_board->WhiteTerritoryBegin();
    while(iter!=m_board->WhiteTerritoryEnd()){
        cr->save();
        cr->set_source_rgb(0,0,0);
        cr->move_to(x0+(iter->i()+0.15)*m_line_space,y0+iter->j()*m_line_space);
        cr->arc(x0+iter->i()*m_line_space,y0+iter->j()*m_line_space,0.15*m_line_space,0, 2*M_PI);
        cr->stroke_preserve();
        cr->set_source_rgb(1,1,1);
        cr->fill();
        cr->restore();
        ++iter;
    }
}

void GobanWidget::draw_ko(Glib::RefPtr<Gdk::GC> gc,Cairo::RefPtr<Cairo::Context> cr)
{
    if(m_board->GetKo()!=Coord(-1,-1)){
        cr->save();
        int x0,y0;
        top_left_corner(x0,y0);
        cr->set_source_rgb(0,0,0);
        cr->move_to(x0+(m_board->GetKo().i()+0.3)*m_line_space,y0+m_board->GetKo().j()*m_line_space);
        cr->arc(x0+m_board->GetKo().i()*m_line_space,y0+m_board->GetKo().j()*m_line_space,0.3*m_line_space,0, 2*M_PI);
        cr->stroke();
        cr->restore();
    }
}

bool GobanWidget::on_motion_notify_event(GdkEventMotion* event)
{
    //cout<<"on_motion_notify_event"<<endl;
    if(!m_board) return true;
    int x,y,gx,gy,x0,y0;
    Gdk::ModifierType state;
    Gdk::Rectangle rect;
    Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(get_window());
    
    top_left_corner(x0,y0);

    if (event->is_hint)
        get_window()->get_pointer(x, y, state);
    else {
        x = event->x;
        y = event->y;
        //state = event->state;
    }

    if (event->x-x0 < -0.5*m_line_space ||
        event->x-x0 >= m_line_space*(m_board->Size()-0.5) ||
        event->y-y0 < -0.5*m_line_space ||
        event->y-y0 >= m_line_space*(m_board->Size()-0.5)){
        //pointer outside of the coord zone.

        if(m_cur_x!=-1 && (*m_board)(m_cur_x,m_cur_y)==EMPTY){
            //erase previous cursor
            rect.set_x((int)(x0+(m_cur_x-0.5)*m_line_space));
            rect.set_y((int)(y0+(m_cur_y-0.5)*m_line_space));
            rect.set_width(m_line_space);
            rect.set_height(m_line_space);
            gc->set_clip_rectangle(rect);
            get_window()->draw_drawable(gc,m_goban_map,0,0,0,0);
        }
        
        m_cur_x=-1;
        m_cur_y=-1;

        return true;
    } else {
        //pointer inside of the coord zone.
        gx=((double)x-x0)/m_line_space + 0.5;
        gy=((double)y-y0)/m_line_space + 0.5;
        //gx,gy goban coords
        if(gx!=m_cur_x || gy!=m_cur_y){
            if(m_cur_x!=-1/* && (*m_board)(m_cur_x,m_cur_y)==EMPTY*/){
                //erase previous cursor
                rect.set_x((int)(x0+(m_cur_x-0.5)*m_line_space));
                rect.set_y((int)(y0+(m_cur_y-0.5)*m_line_space));
                rect.set_width(m_line_space);
                rect.set_height(m_line_space);
                gc->set_clip_rectangle(rect);
                get_window()->draw_drawable(gc,m_goban_map,0,0,0,0);
            }
            if(m_cur_col!=EMPTY){
                if((*m_board)(gx,gy)!=EMPTY){
                    m_cur_x=-1;
                    m_cur_y=-1;
                } else {
                    rect.set_x((int)(x0+(gx-0.5)*m_line_space));
                    rect.set_y((int)(y0+(gy-0.5)*m_line_space));
                    rect.set_width(m_line_space);
                    rect.set_height(m_line_space);
                    gc->set_clip_rectangle(rect);
                    if(m_cur_col==BLACK){
                        get_window()->draw_pixbuf(gc,m_stone_black_cursor,
                                0,0,
                                x0+(gx-0.485)*m_line_space,
                                y0+(gy-0.485)*m_line_space,
                                -1,-1,
                                Gdk::RGB_DITHER_NORMAL,
                                0,0);
                    } else if(m_cur_col==WHITE) {
                        get_window()->draw_pixbuf(gc,m_stone_white_cursor,
                                0,0,
                                x0+(gx-0.485)*m_line_space,
                                y0+(gy-0.485)*m_line_space,
                                -1,-1,
                                Gdk::RGB_DITHER_NORMAL,
                                0,0);
                    }
                    m_cur_x=gx;
                    m_cur_y=gy;
                }
            } else {
                rect.set_x((int)(x0+(gx-0.5)*m_line_space));
                rect.set_y((int)(y0+(gy-0.5)*m_line_space));
                rect.set_width(m_line_space);
                rect.set_height(m_line_space);
                gc->set_clip_rectangle(rect);
                get_window()->draw_pixbuf(gc,m_void_cursor,
                                          0,0,
                                          x0+(gx-0.485)*m_line_space,
                                          y0+(gy-0.485)*m_line_space,
                                          -1,-1,
                                          Gdk::RGB_DITHER_NORMAL,
                                          0,0);
                m_cur_x=gx;
                m_cur_y=gy;
            }
        }
    }
    return true;
}

bool GobanWidget::on_leave_notify_event(GdkEventCrossing* event)
{
    m_cur_x=-1;
    m_cur_y=-1;
    Gtk::Allocation allocation = get_allocation();
    queue_draw();
}

bool GobanWidget::on_button_press_event(GdkEventButton* event)
{
    if(m_board && event->button==1){
        int x,y,x0,y0;
        top_left_corner(x0,y0);

        grab_focus();
        if (event->x-x0 < -0.5*(int)m_line_space ||
            event->x-x0 > m_line_space*(m_board->Size()-0.5) ||
            event->y-y0 < -0.5*(int)m_line_space ||
            event->y-y0 > m_line_space*(m_board->Size()-0.5))
        {
            return true;
        }
        x=(event->x-x0)/m_line_space + 0.5;
        y=(event->y-y0)/m_line_space + 0.5;
        m_signal_stone_dropped.emit(Coord(x,y));
        return true;
    } else
        return false;
}

