#ifndef _GOBANWINDOW_H
#define _GOBANWINDOW_H

#include "GobanWidget.h"
#include "GobanLogic.h"
#include "GamePropertiesDialog.h"
#include "AskPlayerDialog.h"
#include "OpenGameDialog.h"
#include "SgfRecordOpponent.h"
#include "GnugoOpponent.h"

//! Derived class of Gtk::Window to display the main gui interface.
class GobanWindow: public Gtk::Window
{
public:
    //!Constructor for Gtk::Builder
    GobanWindow(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& refBuilder);
    //!Destructor
    /*!
     * The main window owns a GobanLogic instance when a game is opened.
     */
    virtual ~GobanWindow();

private:
    void on_stone_dropped(const Coord& coord);
    void on_new_game();
    void on_close_game();
    void on_quit();
    void on_save_as();
    void on_save();
    void on_open();
    void on_next();
    void on_previous();
    void on_go_beginning();
    void on_go_last();
    void on_next_branch();
    void on_previous_branch();
    void on_game_properties();
    void on_about();
    void on_mode_changed();
    void on_edit_changed();
    void on_comment_changed();
    virtual bool on_delete_event(GdkEventAny* event);
    void on_node_clicked();
    void on_show_ticks();
    void on_show_comments();
    void on_show_next_nodes();
    void on_show_stats();
    void on_pass();
    void on_resign();
    void on_opponent_played(SgfNode* node);
    void on_play_against_gnugo();
    void on_game_over();
    void on_final_state(vector<Coord> dead, vector<Coord> black_terr, vector<Coord> white_terr);
    virtual bool on_button_press_event(GdkEventButton* event);

    bool safe_to_delete_game();
    ustring humancoord(Coord c);
    void actualize_ui();
    void actualize_node();

    Glib::RefPtr<Gtk::Builder>          m_refBuilder;
    Glib::RefPtr<Gtk::ActionGroup>      m_action_group;
    Glib::RefPtr<Gtk::UIManager>        m_ui_manager;
    Gtk::Menu*                          m_popup_menu;
    Glib::RefPtr<Gtk::AccelGroup>       m_accel_group;
    Glib::RefPtr<Gio::File>             m_file;
    Glib::RefPtr<Gio::File>             m_folder;
    Glib::RefPtr<Gtk::TextBuffer>       m_comment_buffer;
    Glib::RefPtr<Gtk::TextView>         m_comment_view;
    Glib::RefPtr<Gtk::RadioAction>      m_mode_edit;
    Glib::RefPtr<Gtk::RadioAction>      m_mode_play;
    Glib::RefPtr<Gtk::RadioAction>      m_edit_add_black;
    Glib::RefPtr<Gtk::RadioAction>      m_edit_add_white;
    Glib::RefPtr<Gtk::RadioAction>      m_edit_remove_stone;
    Glib::RefPtr<Gtk::RadioAction>      m_edit_play;
    Glib::RefPtr<Gtk::Statusbar>        m_statusbar;

    sigc::connection                    m_connection_played;
    sigc::connection                    m_connection_final_state;

    bool                                m_human_black;
    bool                                m_opponent_to_play;
    bool                                m_problem;
    GobanWidget*                        m_gobanwidget;
    GobanLogic*                         m_goban_logic;
    Opponent*                           m_opponent;

    class NodeModel : public Gtk::TreeModelColumnRecord
    {
        public:
            NodeModel() {add(m_id); add(m_text); add(m_full_text); add(m_node);}
            Gtk::TreeModelColumn<Glib::ustring> m_text;
            Gtk::TreeModelColumn<Glib::ustring> m_full_text;
            Gtk::TreeModelColumn<int> m_id;
            Gtk::TreeModelColumn<SgfNode*> m_node;
    };
    NodeModel m_node_model;
    Glib::RefPtr<Gtk::ListStore> m_refListStore;
    Glib::RefPtr<Gtk::TreeSelection> m_refselection;
};

#endif
